/*  -------------------------------------------------------------------------
    Loccalc V1.0, written by Marco Bersani, IK2PIH, Stradella,
    Italy, on 24 August 2002.
    Loccalc is a clone of Locator Calculator V2.0, fully re-written
    in C from scratch.
    Loccalc calculates longitude and latitude from a 4 or 6 characters world
    widelocator, calculates distance (kilometers and miles) between two
    locators or calculates a locator from longitude and latitude.
    You will be prompted for all inputs and all inputs are validity checked.
    The calculation of beam heading and distance assumes that Earth is a sphere


    Locator Calculator V2.0 was a free software written in GW-Basic
    by Lars Berg, SM0CCM,  Jarfalla, Sweden on 15 September 1987, and
    released by the author as a public domain program with no restrictions.

    Loccalc is a program that calculates longitude and latitude from a 4 or 6
    characters world wide locator, calculates distance (kilometers and miles)
    between two locators or calculates a locator from longitude and latitude.
    You will be prompted for all inputs and all inputs are validity checked.

    This software should compile with any C-compiler and run under any OS.
    For better results set EARTHRADIUS to earth radius of your location
    (for my location, north Italy, is 6378 km), but this radius should go
    well for all locations at the same latitude of my QTH (45N or S).

    --------------------------------------------------------------------------
    Not Copyrighted 2002, Marco Bersani, bermar2@inwind.it or mbers@libero.it

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (see the file COPYING which accompany
    this source file);  if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ----------------------------------------------------------------------- */





#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<math.h>

#ifndef PI
# define PI 3.141592654
#endif

#define ISLONGITUDE 0
#define ISLATITUDE  1

#define EARTHRADIUS 6.378   /* Earth radius in north Italy (km) */





int checkvalidlocator(char *locator);
void calculatelongitudeandlatitudefromlocator(char *locator, double *lon, double *lat);
void calculatebeamheadinganddistance(double lon1, double lat1, double lon2, double lat2, double *cb, double *di, double *dmi);
void screenheading(void);
void convertcharactertouppercase(char *string);
void calculatelocatorfromlongitudeandlatitude(char *locator, double lon, double lat);
int getdecimaldegreesfrominputstring(char *inputstring, double *result, char t);





main()
{
/* -------------- Main routine --------------------- */


char mylocator[7];
char otherlocator[7];
char inputstring[9];
double mylon,mylat,otherlon,otherlat,cb,di,dmi;
char choice;
int retval;



do
 {screenheading();
  printf("                                     MAIN MENU \n");
  printf("\n--------------------------------------------------------------------------------\n");
  printf("1. BEAM HEADING AND DISTANCE.\n");
  printf("2. CONVERT LONG/LAT TO LOCATOR.\n");
  printf("3. END\n");

  choice=0;

 while((choice!='1') && (choice!='2') && (choice!='3'))
      {printf("\nMAKE YOUR CHOICE > ");
       scanf("%s",inputstring);
       choice=inputstring[0];}





 /* BEAM HEADING AND DISTANCE */

  if(choice=='1')
   {
     screenheading();
     printf("                          BEAM HEADING AND DISTANCE\n");
     printf("\n--------------------------------------------------------------------------------\n");

     do {
        printf("---------------INSERT YOUR LOCATOR------------------\n");
        printf("(TO QUIT: ENTER \"END\" INSTEAD OF YOUR LOCATOR)\nYOUR LOCATOR? ");
        scanf("%s",mylocator);
        convertcharactertouppercase(mylocator);
        retval=checkvalidlocator(mylocator);


        if(retval==0)
            {printf("WRONG LOCATOR !\n\n");}
        }
     while(retval==0);


     if(retval==-1)
      {calculatelongitudeandlatitudefromlocator(mylocator, &mylon, &mylat);
       printf("\nYOUR LOCATOR IS   = %s\nYOUR LONGITUDE IS = %.2lf DEG\nYOUR LATITUDE IS  = %.2lf DEG\n\n\n",mylocator,mylon,mylat);



       do
       {

        do {
	   printf("\n-------------------INSERT A LOCATOR------------------\n");
	   printf("(TO QUIT: ENTER \"END\" INSTEAD OF A LOCATOR)\nLOCATOR? ");
           scanf("%s",otherlocator);
           convertcharactertouppercase(otherlocator);            /* Read locator and convert in upper case */
           retval=checkvalidlocator(otherlocator);


           if(retval==0)
               {printf("WRONG LOCATOR !\n\n");}
           }
        while(retval==0);                          /* Repeat if input is not valid */



        if(retval==-1)
           {calculatelongitudeandlatitudefromlocator(otherlocator, &otherlon, &otherlat);
            calculatebeamheadinganddistance(mylon, mylat, otherlon, otherlat, &cb, &di, &dmi);
            printf("\n************* CALCULATIONS FOR %s *************\n",otherlocator);
            printf("POSITION OF %s\n",otherlocator);
            printf("LONGITUDE = %.2lf DEG\nLATITUDE  = %.2lf DEG\n",otherlon,otherlat);
	    printf("\nBEAM HEADING AND DISTANCE BETWEEN %s",mylocator);
	    printf(" AND %s\n",otherlocator);
            printf("BEAM HEADING = %.1lf DEG\nDISTANCE  = %.1lf km\nDISTANCE  = %.1lf MLS\n",cb,di,dmi);
            printf("***************************************************\n\n");
           }

       }
       while(retval!=-2);    /* Repeat if locator != END */

     }


   }






 /* CONVERT LONG/LAT TO LOCATOR */

  if(choice=='2')
   {
     screenheading();
     printf("                      CONVERT LONGITUDE/LATITUDE TO LOCATOR \n");
     printf("\n--------------------------------------------------------------------------------");
     printf("\n\n");
     printf("INPUT FORMAT FOR LONGITUDE / LATITUDE = Cnnnsdd\n");
     printf("Where - C   = W/E/N/S  and  nnn = degrees.\n");
     printf("        sdd = .dd for decimal degrees.\n");
     printf("        sdd = 'dd for minutes.\n");

     do
       {printf("\nENTER LONGITUDE - C = W for West, E for East.\n");
        printf("> ");
        scanf("%s",inputstring);
        convertcharactertouppercase(inputstring);
        retval=getdecimaldegreesfrominputstring(inputstring, &mylon, ISLONGITUDE);
        if(retval==0)
            {printf("WRONG LONGITUDE !\n\n");}
       }
     while(retval==0);


     do
       {printf("\nENTER LATITUDE - C = N for North, S for South.\n");
        printf("> ");
        scanf("%s",inputstring);
        convertcharactertouppercase(inputstring);
        retval=getdecimaldegreesfrominputstring(inputstring, &mylat, ISLATITUDE);

        if(retval==0)
            {printf("WRONG LATITUDE !\n\n");}
       }
      while(retval==0);


     calculatelocatorfromlongitudeandlatitude(mylocator, mylon, mylat);
     printf("\nYOUR LOCATOR IS   = %s\nYOUR LONGITUDE IS = %.2lf DEG\nYOUR LATITUDE IS  = %.2lf DEG\n\n",mylocator,mylon,mylat);
   }


 }
while(choice!='3');

return(0);
}









int checkvalidlocator(char *locator)
{
/* -------------- Subroutine -----------------------
      Check valid locator
      Input : char *locator = 4 or 6 characters word wide locator.
      returned value ==  -1 No error. (Valid locator).
      returned value ==  0 Error.   (Invalid locator).
      Note: also string "END" is considered a valid locator, but returned value is -2.
   ------------------------------------------------- */
int lenghtoflocator,x,i;
char t;

lenghtoflocator = strlen(locator);


if (lenghtoflocator == 4)
            {strcat(locator,"LM");
             lenghtoflocator=6;}                 /* Adjust to middle of square */


x=-1;                                                /* Assumes locator is right */


if (strcmp(locator,"END")==0) x=-2;                  /* Return -2 if locator=END */

else if (lenghtoflocator != 6) x=0;                /* force to 0 if locator is wrong */

else for(i=0; i<=5; i++)
       {t=locator[i];
        if(((t > 'R') || (t < 'A')) && ((i==0) || (i==1))) x = 0;  /* first 2 characters: from A to R */
        if(((t > '9') || (t < '0')) && ((i==2) || (i==3))) x = 0;  /* check the numbers  */
        if(((t > 'X') || (t < 'A')) && ((i==4) || (i==5))) x = 0;  /* last 2 characters: from A to X */
       }

return(x);
}













void calculatelongitudeandlatitudefromlocator(char *locator, double *lon, double *lat)
{
/* -------------- Subroutine -----------------------
        calculate longitude and latitude
        Input : locator   = 6 characters world wide locator.
        Output: lon = longitude in decimal degrees;
                lat  = latitude in decimal degrees;
------------------------------------------------- */


*lon = -180+((double)locator[0]-65)*20 + ((double)locator[2]-48)*2 +((double)locator[4]-64.5)/12;
*lon = -(*lon);
*lat = -90+((double)locator[1]-65)*10 + ((double)locator[3]-48) +((double)locator[5]-64.5)/24;
return;
}












void calculatebeamheadinganddistance(double lon1, double lat1, double lon2, double lat2, double *cb, double *di, double *dmi)
{
/* ------Subroutine--------------------
        Calculate beam heading and distance
   Input : lo1  = Longitude 1 in radians.
         : la1  = Latitude 1 in radians.
         : la2  = Latitude 2 in radians.
   Output: cb   = Beam heading in degrees.
           di   = Distance in kilometers.
           dmi  = Distance in miles.
 ------------------------------------ */

double lo1,la1,lo2,la2,x;

lo1=lon1*PI/180;
la1=lat1*PI/180;
lo2=lon2*PI/180;
la2=lat2*PI/180;

(*di)=acos(cos(la1)*cos(lo1)*cos(la2)*cos(lo2)+cos(la1)*sin(lo1)*cos(la2)*sin(lo2)+sin(la1)*sin(la2))*EARTHRADIUS*1000;
(*dmi)=(*di)/1.609;


x=atan2(sin(lo1-lo2)*cos(la2),cos(la1)*sin(la2)-sin(la1)*cos(la2)*cos(lo1-lo2))/PI*180;


if(x<0)
    (*cb)=x+360;
else
    (*cb)=x;


return;
}








void screenheading(void)
{
/* -------------- Subroutine -----------------------
                SCREEN HEADING
   ------------------------------------------------- */
printf("--------------------------------------------------------------------------------\n");
printf("        * * *   IK2PIH LOCCALC - a clone of SM0CCM LOCATOR CALCULATOR  * * *\n");
return;
}









void convertcharactertouppercase(char *st)
{
int lenght;
int i;

/* ----------------SUBROUTINE-----------------------
       CONVERT CHARACTER TO UPPER CASE
       Input : string = Variable length. Upper and/or lower case.
       Output: string = Variable length. Upper case only.
   ------------------------------------------------- */

lenght=strlen(st);

if(lenght!=0)
 {
  for(i=0;i<lenght;i++)
  {
   if(st[i]>='a' && st[i]<='z')
	 st[i]=st[i]-32;
  }
 }
return;
}










void calculatelocatorfromlongitudeandlatitude(char *locator, double lon, double lat)
{
double lo,la;
int alo,bla,clo,dla,elo,fla;

/* -------------- Subroutine -----------------------
   Calculate locator from longitude and latitude
   Input : lon = Longitude in decimal degrees (+ = West;  - = East).
           lat = Latitude in decimal degrees (+ = North; - = South).
   Output: locator = 6 characters world wide locator.
   ------------------------------------------------- */

lo=(-lon+180)/20;
la=(lat+90)/10;

alo=(int)floor(lo);
bla=(int)floor(la);
lo=(lo-(double)alo)*10;
la=(la-(double)bla)*10;

clo = (int)floor(lo);
dla = (int)floor(la);

elo = (int)floor((lo-(double)clo)*24);
fla = (int)floor((la-(double)dla)*24);

locator[0]=(char)(alo+'A');
locator[1]=(char)(bla+'A');
locator[2]=(char)(clo+'0');
locator[3]=(char)(dla+'0');
locator[4]=(char)(elo+'A');
locator[5]=(char)(fla+'A');
locator[6]='\0';

return;
}








int getdecimaldegreesfrominputstring(char *inputstring, double *result, char t)
{
/* -------------- Subroutine -----------------------
      Get decimal degrees from input string
      Return 0 if error occurs; Return -1 if OK;
   ------------------------------------------------- */
char degs[4];
char fractofdegree[3];
char c,d;
double dg,dy;
double decdg;
double mins;
int i;



c=inputstring[0];

if((t==ISLONGITUDE && c!='W' && c!='E') || (t==ISLATITUDE && c!='N' && c!='S'))
                                                                        return(0);
/* Extract degrees from string and convert to double*/
for(i=1;i<=3;i++)
    {if(inputstring[i]>='0' && inputstring[i]<='9') degs[i-1]=inputstring[i];
     else return(0);}
degs[3]='\0';

sscanf(degs,"%lf",&dg);




/* Extract fraction of degrees from string and converts minutes to decimal degrees*/
for(i=5;i<=6;i++)
    {if(inputstring[i]>='0' && inputstring[i]<='9') fractofdegree[i-5]=inputstring[i];
     else return(0);}
fractofdegree[2]='\0';


d=inputstring[4];

if(d=='.')
      sscanf(fractofdegree,"%lf",&decdg);

else if((d=='\'') && (fractofdegree[0]<='5'))
     {sscanf(fractofdegree,"%lf",&mins);
      decdg=mins/60*100;}

else return(0);


/* Build number in decimal degrees and checks result */
dg=dg+(decdg/100);
if((t==ISLONGITUDE && dg>180) || (t==ISLATITUDE && dg>90))
                                                return(0);

/* Minus if longitude east or latitude south */
if(c=='E' || c=='S') {dy=0-dg;
                      dg=dy;}

/* Export result */
*result=dg;

return(-1);
}
